from fastmcp import FastMCP, Image
import matplotlib.pyplot as plt
import os
import seaborn as sns
import time
from typing import Any, Union, Optional
from zaturn import config, query_utils

sns.set_theme()

mcp = FastMCP("Zaturn Visualizations")


def _plot_to_image(plot) -> Union[str, Image]:
    figure = plot.get_figure()
    filepath = os.path.join(config.VISUALS_DIR, str(int(time.time())) + '.png')
    figure.savefig(filepath)
    plt.clf()
    if config.RETURN_IMAGES:
        return Image(path=filepath)
    else:
        return filepath


# Relationships

@mcp.tool()
def scatter_plot(
    query_id: str,
    x: str,
    y: str,
    hue: str = None
    ):
    """
    Make a scatter plot with the dataframe obtained from running SQL Query against source
    If this returns an image, display it. If it returns a file path, mention it.
    Args:
        query_id: Previously run query to use for plotting
        x: Column name from SQL result to use for x-axis
        y: Column name from SQL result to use for y-axis
        hue: Optional String; Column name from SQL result to use for coloring the points  
    """
    df = query_utils.load_query(query_id)
    plot = sns.scatterplot(df, x=x, y=y, hue=hue)
    return _plot_to_image(plot)


@mcp.tool()
def line_plot(
    query_id: str,
    x: str,
    y: str,
    hue: str = None
    ):
    """
    Make a line plot with the dataframe obtained from running SQL Query against source
    Args:
        query_id: Previously run query to use for plotting
        x: Column name from SQL result to use for x-axis
        y: Column name from SQL result to use for y-axis
        hue: Optional; column name from SQL result to use for drawing multiple colored lines
    """
    df = query_utils.load_query(query_id)
    plot = sns.lineplot(df, x=x, y=y, hue=hue)
    return _plot_to_image(plot)


# Distributions

@mcp.tool()
def histogram(
    query_id: str,
    column: str,
    hue: str = None,
    bins: int = None
    ):
    """
    Make a histogram with a column of the dataframe obtained from running SQL Query against source
    Args:
        query_id: Previously run query to use for plotting
        column: Column name from SQL result to use for the histogram
        hue: Optional; column name from SQL result to use for drawing multiple colored histograms
        bins: Optional; number of bins
    """
    df = query_utils.load_query(query_id)
    plot = sns.histplot(df, x=column, hue=hue, bins=bins)
    return _plot_to_image(plot)


# Categorical

@mcp.tool()
def strip_plot(
    query_id: str,
    x: str,
    y: str = None,
    hue: str = None,
    legend: bool = False
    ):
    """
    Make a strip plot with the dataframe obtained from running SQL Query against source
    Args:
        query_id: Previously run query to use for plotting
        x: Column name from SQL result to use for x axis
        y: Optional; column name from SQL result to use for y axis
        hue: Optional; column name from SQL result to use for coloring the points
        legend: Whether to draw a legend for the hue
    """
    df = query_utils.load_query(query_id)
    plot = sns.stripplot(df, x=x, y=y, hue=hue, legend=legend)
    return _plot_to_image(plot)


@mcp.tool()
def box_plot(
    query_id: str,
    x: str,
    y: str = None,
    hue: str = None
    ):
    """
    Make a box plot with the dataframe obtained from running SQL Query against source
    Args:
        query_id: Previously run query to use for plotting
        x: Column name from SQL result to use for x axis
        y: Optional; column name from SQL result to use for y axis
        hue: Optional column name from SQL result to use for coloring the points
    """
    df = query_utils.load_query(query_id)
    plot = sns.boxplot(df, x=x, y=y, hue=hue)
    return _plot_to_image(plot)


@mcp.tool()
def bar_plot(
    query_id: str,
    x: str,
    y: str = None,
    hue: str = None,
    orient: str = 'v'
    ):
    """
    Make a bar plot with the dataframe obtained from running SQL Query against source
    Args:
        query_id: Previously run query to use for plotting
        x: Column name from SQL result to use for x axis
        y: Optional; column name from SQL result to use for y axis
        hue: Optional column name from SQL result to use for coloring the bars
        orient: Orientation of the box plot, use 'v' for vertical and 'h' for horizontal
    """
    df = query_utils.load_query(query_id)
    plot = sns.barplot(df, x=x, y=y, hue=hue, orient=orient)
    return _plot_to_image(plot)





import argparse
import os
import platformdirs
import pkg_resources
import sys

# Basic Setup
USER_DATA_DIR = platformdirs.user_data_dir('zaturn', 'zaturn')
USER_CONFIG_DIR = platformdirs.user_config_dir('zaturn', 'zaturn')
QUERIES_DIR = os.path.join(USER_DATA_DIR, 'queries')
VISUALS_DIR = os.path.join(USER_DATA_DIR, 'visuals')
SOURCES_FILE = os.path.join(USER_CONFIG_DIR, 'sources.txt')

os.makedirs(USER_CONFIG_DIR, exist_ok=True)
os.makedirs(QUERIES_DIR, exist_ok=True)
os.makedirs(VISUALS_DIR, exist_ok=True)

# Parse command line args
parser = argparse.ArgumentParser(
    description="Zaturn: A read-only BI tool for analyzing various data sources"
)
parser.add_argument('--noimg', action='store_const', 
    const=True, default=False, 
    help='Return image file paths instead of images for visuals. Use when MCP client cannot render images.',
)
parser.add_argument('sources', nargs=argparse.REMAINDER, default=[], 
    help='Data source (can be specified multiple times). Can be SQLite, MySQL, PostgreSQL connection string, or a path to CSV, Parquet, or DuckDB file.'
)
args = parser.parse_args()

# Read and parse sources
source_list = []
if os.path.exists(SOURCES_FILE):
    with open(SOURCES_FILE) as f:
        source_list = [line.strip('\n') for line in f.readlines() if line.strip('\n')]

if not source_list:
    source_list = args.sources

if not source_list:
    source_list = [
        pkg_resources.resource_filename(
            'zaturn',
            os.path.join('example_data', 'all_pokemon_data.csv')
        )
    ]
    print("No data sources provided. Loading example dataset for demonstration.")
    print(f"\nTo load your datasets, add them to {SOURCES_FILE} (one source URL or full file path per line)")
    print("\nOr use command line args to specify data sources:")
    print("zaturn_mcp sqlite:///path/to/mydata.db /path/to/my_file.csv")
    print(f"\nNOTE: Sources in command line args will be ignored if sources are found in {SOURCES_FILE}")
    
SOURCES = {}
for s in source_list:
    source = s.lower()
    if source.startswith('sqlite://'):
        source_type = 'sqlite'
        source_name = source.split('/')[-1].split('?')[0].split('.db')[0]
    elif source.startswith('postgresql://'):
        source_type = 'postgresql'
        source_name = source.split('/')[-1].split('?')[0]
    elif source.startswith("mysql+pymysql://"):
        source_type = 'mysql'
        source_name = source.split('/')[-1].split('?')[0]
    elif source.endswith(".duckdb"):
        source_type = "duckdb"
        source_name = source.split('/')[-1].split('.')[0]
    elif source.endswith(".csv"):
        source_type = "csv"
        source_name = source.split('/')[-1].split('.')[0]
    elif source.endswith(".parquet"):
        source_type = "parquet"
        source_name = source.split('/')[-1].split('.')[0]
    else:
        continue

    source_id = f'{source_name}-{source_type}'
    if source_id in SOURCES:
        i = 2
        while True:
            source_id = f'{source_name}{i}-{source_type}'
            if source_id not in SOURCES:
                break
            i += 1

    SOURCES[source_id] = {'url': s, 'type': source_type}


# Other Settings
RETURN_IMAGES = not args.noimg






from fastmcp import FastMCP
from zaturn import core, visualizations

# Mount modules and make MCP
mcp = FastMCP("Zaturn MCP")
mcp.mount("core", core.mcp)
mcp.mount("visualizations", visualizations.mcp)

def main():
    mcp.run()


if __name__=="__main__":
    main()



import duckdb
import numpy as np
import os
import pandas as pd
import sqlalchemy
from sqlalchemy.orm import Session
import time
from typing import List
from zaturn import config


def execute_query(source: dict, query: str):
    """Run the query using the appropriate engine and read only config"""
    url = source['url']
    
    match source['type']:
        case "sqlite":
            if "mode=ro" in url:
                pass
            elif '?' in url:
                url += '&mode=ro'
            else:
                url += '?mode=ro'
            with sqlalchemy.create_engine(url).connect() as conn:
                result = conn.execute(sqlalchemy.text(query))
                return pd.DataFrame(result)

        case "mysql":
            engine = sqlalchemy.create_engine(url)
            with Session(engine) as session:
                session.autoflush = False
                session.autocommit = False
                session.flush = lambda *args: None
                result = session.execute(sqlalchemy.text(query))
                return pd.DataFrame(result)

        case "postgresql":
            engine = sqlalchemy.create_engine(url)
            with engine.connect() as conn:
                conn = conn.execution_options(
                    isolation_level="SERIALIZABLE",
                    postgresql_readonly=True,
                    postgresql_deferrable=True,
                )
                with conn.begin():
                    result = conn.execute(sqlalchemy.text(query))
                    return pd.DataFrame(result)

        case "duckdb":
            conn = duckdb.connect(url, read_only=True)
            return conn.execute(query).df()

        case "csv":
            conn = duckdb.connect(database=':memory:')
            conn.execute(f"CREATE VIEW CSV AS SELECT * FROM read_csv('{url}')")
            return conn.execute(query).df()

        case "parquet":
            conn = duckdb.connect(database=':memory:')
            conn.execute(f"CREATE VIEW PARQUET AS SELECT * FROM read_parquet('{url}')")
            return conn.execute(query).df()
        
        case _:
            raise Exception("Unsupported Source")


def save_query(df: pd.DataFrame):
    """Save query results to disk and return a unique reference id"""
    query_id = 'q' + str(int(time.time()))
    filepath = os.path.join(config.QUERIES_DIR, f'{query_id}.parquet')
    df.replace({np.nan: None}).to_parquet(filepath, engine='pyarrow', index=False)
    return query_id


def load_query(query_id: str):
    """Load query results from disk using unique reference id"""
    filepath = os.path.join(config.QUERIES_DIR, f'{query_id}.parquet')
    df = pd.read_parquet(filepath, engine='pyarrow').replace({np.nan: None})
    df.reset_index(drop=True, inplace=True)
    return df



from fastmcp import FastMCP
import os
from typing import Any, List, Union
from zaturn import config, query_utils

mcp = FastMCP("Zaturn Core")


@mcp.tool()
def list_sources() -> str:
    """
    List all available data sources.
    Returns a list of unique source_ids to be used for other queries.
    Source type is included in the source_id string.
    While drafting SQL queries use appropriate syntax as per source type.
    """
    try:
        if not config.SOURCES:
            return "No data sources available. Add sources using the command line parameters."
        
        result = "Available data sources:\n\n"
        for source in config.SOURCES:
            result += f"- {source}\n"
    
        return result
    except Exception as e:
        return str(e)


@mcp.tool()
def list_tables(source_id: str):
    """
    Lists names of all tables/datasets in a given data source.
    Use run_query with appropriate SQL query to determine table structure
    
    Args:
        source_id: The data source to list tables from
    """
    try:
        source = config.SOURCES.get(source_id)
        if not source:
            return f"Source {source_id} Not Found"

        match source['type']:
            case "sqlite":
                result = query_utils.execute_query(source,
                    "SELECT name FROM sqlite_schema WHERE type ='table' AND name NOT LIKE 'sqlite_%';"
                )
                return result['name'].to_list()

            case "postgresql":
                result = query_utils.execute_query(source,
                    "SELECT tablename FROM pg_catalog.pg_tables WHERE schemaname != 'pg_catalog' AND schemaname != 'information_schema';"
                )
                return result['tablename'].to_list()

            case "mysql":
                result = query_utils.execute_query(source, "SHOW TABLES")
                for col in list(result):
                    if col.startswith("Tables_in_"):
                        return result[col].to_list()
                
            case "duckdb" | "csv" | "parquet":
                result = query_utils.execute_query(source, "SHOW TABLES")
                return result['name'].to_list()

    except Exception as e:
        return str(e)

@mcp.tool()
def run_query(source_id: str, query: str) -> str:
    """
    Run query against specified source
    For both csv and parquet sources, use DuckDB SQL syntax
    Use 'CSV' as the table name for csv sources.
    Use 'PARQUET' as the table name for parquet sources.

    This will return a query_id, which can be referenced while calling other Zaturn tools.
    Args:
        source_id: The data source to run the query on
        query: SQL query to run on the data source
    """
    try:
        source = config.SOURCES.get(source_id)
        if not source:
            return f"Source {source_id} Not Found"
            
        df = query_utils.execute_query(source, query)
        query_id = query_utils.save_query(df)
        return query_id
    except Exception as e:
        return str(e)


@mcp.tool()
def show_query_result(query_id) -> str:
    """
    Show stored result for query_id in markdown table format
    """
    df = query_utils.load_query(query_id)
    return df.to_markdown(index=False)
